
#include "contour_dual.h"



extern Vect3f point_non_lie;

/*********************************************************************
                Routines de fonctionnement du contour dual
**********************************************************************/

Contour_dual::Contour_dual(Pot0 *p,float ref,int profondeur_detection,int profondeur,float largeurmax):
    oct(),trac(oct,largeurmax/POSMAX,0.5*largeurmax*Vect3f(1,1,1)),cube_manquant(100){
    potblob=p;
    potentiel_ref=ref;
    e_feuille=EMAX-profondeur;

    detection(EMAX -profondeur_detection);
    triangule();
    amelioration();
    calculenormale();


    //Génération des buffer
    glGenBuffers(2,buffer);
    // Buffer d'informations de vertex
    glBindBuffer(GL_ARRAY_BUFFER, buffer[0]);
    glBufferData(GL_ARRAY_BUFFER, tabsommets.endalloc()*sizeof(Vertexcolornormal), NULL, GL_DYNAMIC_DRAW);
    // Buffer d'indices
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffer[1]);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, tabtriangles.endalloc()*sizeof(Triangle), NULL, GL_DYNAMIC_DRAW);
    // On rentre les données dans le buffer
    subdata();
}

void Contour_dual::detection(int e_grille){
    (*trac.gethead()).splitrec(EMAX-e_grille-1);
    trac.firstlastcube();do{
        Vect3f p0 = trac.getcoordonne();
        float d0 = trac.getlargeur();

        while(trac.getexp()>e_feuille){
            // On subdivise si le cube a deux coin de signes différents
            // Ce qui équivaut à ce qu'il existe un coin qui soit de signe opposé au premier coin
            bool signe0=(*potblob)(p0)>potentiel_ref;
            if(    (signe0 != ((*potblob)(p0+Vect3f(d0,0,0))>potentiel_ref))
                || (signe0 != ((*potblob)(p0+Vect3f(0,d0,0))>potentiel_ref))
                || (signe0 != ((*potblob)(p0+Vect3f(d0,d0,0))>potentiel_ref))
                || (signe0 != ((*potblob)(p0+Vect3f(0,0,d0))>potentiel_ref))
                || (signe0 != ((*potblob)(p0+Vect3f(d0,0,d0))>potentiel_ref))
                || (signe0 != ((*potblob)(p0+Vect3f(0,d0,d0))>potentiel_ref))
                || (signe0 != ((*potblob)(p0+Vect3f(d0,d0,d0))>potentiel_ref))
               ){
                trac.split_go();// On subdivise et on se place au premier fils
                d0*=0.5;
            }else break;
        }
    }while(trac.nextlastcube());
}


void Contour_dual::select(Traceur<Voxel_contour> const&trac0){
    if(!trac0.getnode().deja_teste){
    float l=trac0.getscale()*(1<<e_feuille);
    Vect3f p0 = trac0.getcoordonne()+0.5*l*Vect3f(1,1,1);// les coordonnées du centre du voxel
    int d =1<<e_feuille;
    int px=trac0.getposxyz()[0],py=trac0.getposxyz()[1],pz=trac0.getposxyz()[2];
    int degre=0;


    bool u0=(*potblob)(p0)<=potentiel_ref;
    bool u1=(*potblob)(p0+Vect3f(l,0,0))<=potentiel_ref;
    bool u2=(*potblob)(p0+Vect3f(0,l,0))<=potentiel_ref;
    bool u3=(*potblob)(p0+Vect3f(0,0,l))<=potentiel_ref;
    bool u4=(*potblob)(p0+Vect3f(-l,0,0))<=potentiel_ref;
    bool u5=(*potblob)(p0+Vect3f(0,-l,0))<=potentiel_ref;
    bool u6=(*potblob)(p0+Vect3f(0,0,-l))<=potentiel_ref;

    if(u0)degre++;
    if(u1)degre++;
    if(u2)degre++;
    if(u3)degre++;
    if(u4)degre++;
    if(u5)degre++;
    if(u6)degre++;


    if(u0 && degre<=6){

        //On allume le cube en sélectionnant les faces libres
        trac0.getnode().faces_libres=0;
        if(!u1)trac0.getnode().faces_libres|=8;
        if(!u2)trac0.getnode().faces_libres|=16;
        if(!u3)trac0.getnode().faces_libres|=32;
        if(!u4)trac0.getnode().faces_libres|=1;
        if(!u5)trac0.getnode().faces_libres|=2;
        if(!u6)trac0.getnode().faces_libres|=4;

        //On vérifie qu'il n'y ai pas d'erreur topologiques
        //*
        if(!u1 && !u2  && (*potblob)(p0+Vect3f(l,l,0)) <= potentiel_ref)
            set_collapsed(px+d,py+d,pz,trac,px+d,py+d,pz+d);
        if(!u1 && !u3  && (*potblob)(p0+Vect3f(l,0,l)) <= potentiel_ref)
            set_collapsed(px+d,py,pz+d,trac,px+d,py+d,pz+d);
        if(!u2 && !u3  && (*potblob)(p0+Vect3f(0,l,l)) <= potentiel_ref)
            set_collapsed(px,py+d,pz+d,trac,px+d,py+d,pz+d);

        if(!u4 && !u2  && (*potblob)(p0+Vect3f(-l,l,0)) <= potentiel_ref)
            set_collapsed(px,py+d,pz,trac,px,py+d,pz+d);
        if(!u4 && !u3  && (*potblob)(p0+Vect3f(-l,0,l)) <= potentiel_ref)
            set_collapsed(px,py,pz+d,trac,px,py+d,pz+d);
        if(!u5 && !u3  && (*potblob)(p0+Vect3f(0,-l,l)) <= potentiel_ref)
            set_collapsed(px,py,pz+d,trac,px+d,py,pz+d);
        //*/

        }
        //On regarde les cubes à côté
        //*
        if(degre>=1 && degre<=6){
            if(trac.pwriteoct(px+d,py,pz,e_feuille)!=NULL && !trac.getnode().sera_teste){
                cube_manquant.add(Vect3i(px+d,py,pz));trac.getnode().sera_teste=true;}
            if(trac.pwriteoct(px,py+d,pz,e_feuille)!=NULL && !trac.getnode().sera_teste){
                cube_manquant.add(Vect3i(px,py+d,pz));trac.getnode().sera_teste=true;}
            if( trac.pwriteoct(px,py,pz+d,e_feuille)!=NULL && !trac.getnode().sera_teste){
                cube_manquant.add(Vect3i(px,py,pz+d));trac.getnode().sera_teste=true;}

            if(trac.pwriteoct(px-d,py,pz,e_feuille)!=NULL && !trac.getnode().sera_teste){
                cube_manquant.add(Vect3i(px-d,py,pz));trac.getnode().sera_teste=true;}
            if(trac.pwriteoct(px,py-d,pz,e_feuille)!=NULL && !trac.getnode().sera_teste){
                cube_manquant.add(Vect3i(px,py-d,pz));trac.getnode().sera_teste=true;}
            if( trac.pwriteoct(px,py,pz-d,e_feuille)!=NULL && !trac.getnode().sera_teste){
                cube_manquant.add(Vect3i(px,py,pz-d));trac.getnode().sera_teste=true;}

        }//*/
        trac0.getnode().sera_teste=true;
    }
    trac0.getnode().deja_teste=true;
}

void Contour_dual::triangule(){
    Traceur<Voxel_contour> trac0(trac);


    trac0.firstlastcube();do{
        trac0.getnode().sera_teste=true;// les voxels seront testé car l'on va parcourir l'octree
        trac0.getnode().deja_teste=false;
    }while(trac0.nextlastcube());


    trac0.firstlastcube();do{//on place les indices dans les voxels en parcourant l'octree
        if(e_feuille==trac0.getexp())
            select(trac0);
    }while(trac0.nextlastcube());

    while(cube_manquant.end()>0){// On propage le front de Voxels en vidant le tableau cube_manquant
        Vect3i id=cube_manquant.first();
        trac0.getcube(id.x,id.y,id.z,e_feuille);
        select(trac0);
        cube_manquant.indice_suppr(0);
    }
    //*
    int d =1<<e_feuille;
    trac0.firstlastcube();do{//on maille les quadrangles (6 faces par voxel au maximum)
        if(trac0.getnode().faces_libres>0){
        int ifaces =trac0.getnode().faces_libres;
        int ic1,ic2,ic3,ic4;
        int px=trac0.getposxyz()[0],py=trac0.getposxyz()[1],pz=trac0.getposxyz()[2];

            if(ifaces & 8){
                ic1=getindicecoin(px+d,py,pz);// On cherche l'indice du coin dans le voxel dans lequel il doit être rangé
                ic2=getindicecoin(px+d,py,pz+d);
                ic3=getindicecoin(px+d,py+d,pz+d);
                ic4=getindicecoin(px+d,py+d,pz);
                meshquadrangle(ic1,ic2,ic3,ic4);
            }
            if(ifaces & 16){
                ic1=getindicecoin(px,py+d,pz);
                ic2=getindicecoin(px+d,py+d,pz);
                ic3=getindicecoin(px+d,py+d,pz+d);
                ic4=getindicecoin(px,py+d,pz+d);
                meshquadrangle(ic1,ic2,ic3,ic4);
            }
            if(ifaces & 32){
                ic1=getindicecoin(px,py,pz+d);
                ic2=getindicecoin(px+d,py,pz+d);
                ic3=getindicecoin(px+d,py+d,pz+d);
                ic4=getindicecoin(px,py+d,pz+d);
                meshquadrangle(ic1,ic2,ic3,ic4);
            }
            if(ifaces & 1){
                ic1=getindicecoin(px,py,pz);
                ic2=getindicecoin(px,py,pz+d);
                ic3=getindicecoin(px,py+d,pz+d);
                ic4=getindicecoin(px,py+d,pz);
                meshquadrangle(ic1,ic2,ic3,ic4);
            }
            if(ifaces & 2){
                ic1=getindicecoin(px,py,pz);
                ic2=getindicecoin(px+d,py,pz);
                ic3=getindicecoin(px+d,py,pz+d);
                ic4=getindicecoin(px,py,pz+d);
                meshquadrangle(ic1,ic2,ic3,ic4);
            }
            if(ifaces & 4){
                ic1=getindicecoin(px,py,pz);
                ic2=getindicecoin(px+d,py,pz);
                ic3=getindicecoin(px+d,py+d,pz);
                ic4=getindicecoin(px,py+d,pz);
                meshquadrangle(ic1,ic2,ic3,ic4);
            }
        }
    }while(trac0.nextlastcube());//*/
}

int Contour_dual::getindicecoin(int px,int py,int pz){
    if(trac.pwriteoct(px,py,pz,e_feuille)!=NULL){
        if(trac.getnode().indice_coin<0){// si le sommet n'existe pas on le crée
            trac.getnode().indice_coin=tabsommets.end();
            tabsommets.add(Vertexcolornormal(trac.getcoordonne()));
            plaquevect(tabsommets.last());
        }
        return trac.getnode().indice_coin;// on rend le numero du sommet
    }else return -1;// rend -1 si l'on est hors de l'octree
}
void Contour_dual::set_collapsed(int px1,int py1,int pz1,Traceur<Voxel_contour> &trac2,int px2,int py2,int pz2){
    //On vérifie que les sommets ne sont pas déjà fusionné, pour cela il suffit de vérifier s'il sont définis,
    //puis on fusionne les sommets.
    if(trac2.pwriteoct(px1,py1,pz1,e_feuille)!=NULL && trac2.getnode().indice_coin<0){
            int &n1 =trac2.getnode().indice_coin;
            Vect3f v1=trac2.getcoordonne();
            if(trac2.pwriteoct(px2,py2,pz2,e_feuille)!=NULL && trac2.getnode().indice_coin<0){
                n1=tabsommets.end();
                trac2.getnode().indice_coin=tabsommets.end();
                tabsommets.add(0.5*(v1+trac2.getcoordonne()));// On place le nouveau point au milieu du segment
                plaquevect(tabsommets.last());
        }
    }
}

void Contour_dual::meshquadrangle(int a,int b,int c,int d){
    if(a>-1 && b> -1 && c>-1 && d>-1){// vérifie que les sommets sont bien définis
        Vect3f v0 = tabsommets(a)-tabsommets(c);
        Vect3f v1 = tabsommets(b)-tabsommets(d);
        if(v0*v0>v1*v1){// crée 2 triangles pour former le quadrangle en prenant la diagonale la plus courte
            if(a!=b && b!=d && d!=a)
            tabtriangles.add(Triangle(a,b,d));
            if(c!=b && b!=d && d!=c)
            tabtriangles.add(Triangle(c,b,d));
        }else{
            if(a!=b && b!=c && c!=a)
            tabtriangles.add(Triangle(b,a,c));
            if(a!=d && d!=c && c!=a)
            tabtriangles.add(Triangle(d,a,c));
        }
    }
}

void Contour_dual::amelioration(){
    int k,a,b,c;
    Vect3f Gk;
    Vect3f *sommets = new Vect3f[tabsommets.end()];
    int *cardinal = new int[tabsommets.end()];
    for(k=0;k<tabsommets.end();k++)cardinal[k]=0;

    // On calcule le barycentre des faces  des triangles autours de chaque sommet
    for(k=0;k<tabtriangles.end();k++){
        a=tabtriangles(k).a;
        b=tabtriangles(k).b;
        c=tabtriangles(k).c;
        Gk=tabsommets(a)+tabsommets(b)+tabsommets(c);
        sommets[a]+=Gk;
        sommets[b]+=Gk;
        sommets[c]+=Gk;
        cardinal[a]+=3;
        cardinal[b]+=3;
        cardinal[c]+=3;
    }
    // On reprojette les sommets
    for(k=0;k<tabsommets.end();k++){
        sommets[k]*=1./(float)cardinal[k];
        plaquevect(sommets[k]);
        (Vect3f&)tabsommets(k)=sommets[k];// On replace le résultat dans le tableau initial
    }
    delete[] sommets;
    delete[] cardinal;
}

/*******************************************************************************
            Fonctions pour l'affichages du contour
********************************************************************************/

void Contour_dual::subdata(){// mise à jour du VBO
    glBindBuffer(GL_ARRAY_BUFFER, buffer[0]);
    glBufferSubData(GL_ARRAY_BUFFER, 0, tabsommets.end()*sizeof(Vertexcolornormal), (GLfloat*)&tabsommets.first());
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffer[1]);
    glBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 0, tabtriangles.end()*sizeof(Triangle), (GLint*)&tabtriangles.first());
}

void Contour_dual::affiche(){
    glBindBuffer(GL_ARRAY_BUFFER, buffer[0]);
    glVertexPointer( 3, GL_FLOAT,sizeof(Vertexcolornormal),0);
    glColorPointer( 4, GL_UNSIGNED_BYTE, sizeof(Vertexcolornormal), (float*)NULL +(3) );
    glNormalPointer(GL_FLOAT, sizeof(Vertexcolornormal), (float*)NULL +(4) );

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffer[1]);
    // Activation d'utilisation des tableaux
    glEnableClientState( GL_VERTEX_ARRAY );
    glEnableClientState( GL_COLOR_ARRAY );
    glEnableClientState( GL_NORMAL_ARRAY );
    // Rendu de notre géométrie
    glDrawElements(GL_TRIANGLES, tabtriangles.end()*3, GL_UNSIGNED_INT, 0);

    glDisableClientState( GL_NORMAL_ARRAY );
    glDisableClientState( GL_COLOR_ARRAY );
    glDisableClientState( GL_VERTEX_ARRAY );
}

void Contour_dual::setcolor(Couleurs &c){
    int k;
    for(k=0;k<tabsommets.end();k++){
        tabsommets(k).color=c(tabsommets(k));
    }
    subdata();
}
void Contour_dual::setcolor(int c){
    int k;
    for(k=0;k<tabsommets.end();k++){
        tabsommets(k).color=c;
    }
    subdata();
}

void Contour_dual::calculenormale(){
    int k;
    for(k=0;k<tabsommets.end();k++){
        tabsommets(k).normale=-getgrad(tabsommets(k));
        //renorme le gradient
        tabsommets(k).normale*=1./sqrt(tabsommets(k).normale*tabsommets(k).normale);
    }
}
Toile Contour_dual::gettoile(int couleur){
    int *indices = new int[tabtriangles.end()*6];
    Vertexcolornormal *vertex= new Vertexcolornormal[tabsommets.end()];
    int k;
    for(k=0;k<tabsommets.end();k++){
        vertex[k]=Vertexcolornormal(tabsommets(k));
        vertex[k].color=couleur;
    }
    for(k=0;k<tabtriangles.end();k++){
        indices[k*6] = tabtriangles(k).a;
        indices[k*6+1] = tabtriangles(k).b;
        indices[k*6+2] = tabtriangles(k).b;
        indices[k*6+3] = tabtriangles(k).c;
        indices[k*6+4] = tabtriangles(k).a;
        indices[k*6+5] = tabtriangles(k).c;
    }
    return Toile(vertex,tabsommets.end(),indices,tabtriangles.end()*3);
}

void Contour_dual::printstats(){
    printf("statistiques du contour :\n");
    printf("    nombre de vertex %d ; nombre de triangles %d \n",
           tabsommets.end(),tabtriangles.end());
    int k;
    float airemin=100000.;
    double dmoyen=0;
    float dmin=100.;
    float dmax=0.;
    for(k=0;k<tabtriangles.end();k++){
        Vect3f p1=tabsommets(tabtriangles(k).a);
        Vect3f p2=tabsommets(tabtriangles(k).b);
        Vect3f p3=tabsommets(tabtriangles(k).c);

        Vect3f v=(p1-p2)^(p1-p3);
        airemin=min((float)sqrt(v*v),airemin);

        float d2 = sqrt((p1-p2)*(p1-p2));
        dmin=min(dmin,d2);
        dmax=max(dmax,d2);
        dmoyen+=(double)d2;

        d2 = sqrt((p1-p3)*(p1-p3));
        dmin=min(dmin,d2);
        dmax=max(dmax,d2);
        dmoyen+=(double)d2;

        d2 = sqrt((p3-p2)*(p3-p2));
        dmin=min(dmin,d2);
        dmax=max(dmax,d2);
        dmoyen+=(double)d2;
    }
    dmoyen*=1./(3.*(double)tabtriangles.end());
    dmin*=1./dmoyen;
    dmax*=1./dmoyen;
    airemin*=1./carre(dmoyen);
    // Les distances relatives sont relatives à la distance moyenne
    printf("    distance moyenne %f \n    aire min relative %f \n    distance min relative %f \n    distance max relative %f \n",
           dmoyen,airemin,dmin,dmax);
}


