#ifndef CONTOUR_DUAL_H
#define CONTOUR_DUAL_H


#include "math.h"
#include "potentiel.h"
#include "matr9f.h"
#include "affichable.h"


class Voxel_contour{
    public:
    bool sera_teste;// sera testé ou a déjà été testé
    bool deja_teste;// a déjà été testé
    int faces_libres;// les 6 premiers bits indiquent pour chaque face s'il faut mailler la face.
    int indice_coin;// l'indice du sommet du coin relatif au voxel, vaut -1 si non défini

    inline Voxel_contour(){
        sera_teste=false;
        deja_teste=false;
        indice_coin=-1;
        faces_libres=0;
    }
};

// une structure pour stocker des coordonnée entière.
class Vect3i{
    public:

    inline Vect3i(){x=y=z=0;}
    inline Vect3i(int a,int b, int c){x=a;y=b;z=c;}

    int x;
    int y;
    int z;
};

class Contour_dual{
        public:
    dynamictab<Vertexcolornormal> tabsommets;//Le tableau des  vertexs
    dynamictab<Triangle> tabtriangles;//Le tableau des triangles

    Pot0 *potblob;// Un pointeur sur un potentiel
    float potentiel_ref;// Le potentiel de référence

    Octree<Voxel_contour> oct;// L'Octree
    Traceur<Voxel_contour> trac;// Un itérateur pour Octree
    dynamictab<Vect3i> cube_manquant;// On stocke ici les coordonnées des voxels à tester
    int e_feuille;// La taille logarithmique des voxels


    GLuint buffer[2];// le buffer du VBO

    Contour_dual(Pot0 *p,float ref,int profondeur_detection,int profondeur,float largeurmax);
    // le constructeur qui réalise la triangulation avec amélioration, et le calcul de normale

    ~Contour_dual(){glDeleteBuffers(2,buffer);}// le destructeur

    Toile gettoile(int couleur);


    void setcolor(Couleurs &c);
    void setcolor(int c);

    void calculenormale();

    void affiche();
    void subdata();// remplis le VBO (vertex buffer object)

    void triangule();
    //réalise la triangulation

    void amelioration();

    void printstats();
    // affiche les statistiques dans la console, elles ne s'affichent que si l'on lance le programme dans une console

    inline Vect3f getgrad(Vect3f pos){return (*potblob).getgrad(pos);}
    inline void plaquevect(Vect3f &pos){(*potblob).plaquevect(pos,potentiel_ref);}
    //plaque le vecteur sur la surface en utilisant la méthode de Newton

        private:

    void detection(int e_grille);
    //détecte la surface et préaloue des voxels

    void select(Traceur<Voxel_contour> const&trac0);
    // rentre les caractéristiques du voxel qui est pointé par le traceur

    int getindicecoin(int px,int py,int pz);
    // rend l'indice du vertex relatif au voxel de coordonnées (px,py,pz)

    void set_collapsed(int px1,int py1,int pz1,Traceur<Voxel_contour> &trac2,int px2,int py2,int pz2);
    // crée les vertex et les fusionne

    void meshquadrangle(int i1,int i2,int i3,int i4);
    // rentre les triangles dans le tableau à partir de l'indice des sommets du quadrangle
};





#endif
