#include <GL/glew.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "matr9f.h"

Matr9f Vect3f::v()const{
    return Matr9f(Vect3f(0,-z,y),
                  Vect3f(z,0,-x),
                  Vect3f(-y,x,0));
}
Vect3f vectorielle(Vect3f const& a,Vect3f const& b){
    return Vect3f(
        a.y*b.z-a.z*b.y,
        b.x*a.z-b.z*a.x,
        a.x*b.y-a.y*b.x);
}
Vect3f operator^(Vect3f const& a,Vect3f const& b){
    return Vect3f(
        a.y*b.z-a.z*b.y,
        b.x*a.z-b.z*a.x,
        a.x*b.y-a.y*b.x);
}
Vect3f doubleproduit(Vect3f const& a,Vect3f const& b,Vect3f const& c){
    return (a*c)*b-(a*b)*c;
}
Vect3f findorthogonal(Vect3f const&v){
    Vect3f p;
    if(fabs(v.x)<fabs(v.y) && fabs(v.x)<fabs(v.z))p=Vect3f(1,0,0);
    else if(fabs(v.y)<fabs(v.z))p=Vect3f(0,1,0);
    else p=Vect3f(0,0,1);//on prend la plus petite valeur

    p=surfaceprojection(v,p);
    p=1./sqrt(p*p)*p;
    return p;
}
Vect3f surfaceprojection(Vect3f normale, Vect3f u){
    return u-(normale*u)/(normale*normale)*normale;
}
Matr9f::Matr9f(){
    int k;
    for(k=0;k<9;k++)PTR_MATR[k]=0;
}
Matr9f::Matr9f(float d){
    int k;
    for(k=0;k<9;k++)PTR_MATR[k]=0;
    PTR_MATR[0]=d;
    PTR_MATR[4]=d;
    PTR_MATR[8]=d;
}
Matr9f::Matr9f(Vect3f const& diagj){
    int k;
    for(k=0;k<9;k++)PTR_MATR[k]=0;
    PTR_MATR[0]=diagj.x;
    PTR_MATR[4]=diagj.y;
    PTR_MATR[8]=diagj.z;
}
Matr9f::Matr9f(float a,float b, float c){
    int k;
    for(k=0;k<9;k++)PTR_MATR[k]=0;
    PTR_MATR[0]=a;
    PTR_MATR[4]=b;
    PTR_MATR[8]=c;
}
void Matr9f::transpo(){
    echange(Lx.y,Ly.x);
    echange(Lz.y,Ly.z);
    echange(Lz.x,Lx.z);
}
Matr9f transpo(Matr9f const& m){Matr9f a(m);a.transpo();return a;}

Vect3f operator*(Matr9f const& m, Vect3f const& a){
    return Vect3f(m.Lx*a,m.Ly*a,m.Lz*a);
}
void Matr9f::operator+=(Matr9f const& a){Lx+=a.Lx;Ly+=a.Ly;Lz+=a.Lz;}
void Matr9f::operator-=(Matr9f const& a){Lx-=a.Lx;Ly-=a.Ly;Lz-=a.Lz;}
void Matr9f::operator*=(float a){Lx*=a;Ly*=a;Lz*=a;}

Matr9f operator*(float a, Matr9f const& b){
    Matr9f c(b); c*=a; return c;}

Matr9f operator*(Matr9f const& a, Matr9f const& b){
    return Matr9f(a.Lx.x*b.Lx+a.Lx.y*b.Ly+a.Lx.z*b.Lz,
                  a.Ly.x*b.Lx+a.Ly.y*b.Ly+a.Ly.z*b.Lz,
                  a.Lz.x*b.Lx+a.Lz.y*b.Ly+a.Lz.z*b.Lz);
}

Matr9f operator!(Matr9f const& b){
    Matr9f m;
    m[0]=b[4]*b[8]-b[5]*b[7];
    m[1]=b[2]*b[7]-b[8]*b[1];
    m[2]=b[1]*b[5]-b[2]*b[4];
    m[3]=b[5]*b[6]-b[8]*b[3];
    m[4]=b[0]*b[8]-b[6]*b[2];
    m[5]=b[2]*b[3]-b[0]*b[5];
    m[6]=b[3]*b[7]-b[4]*b[6];
    m[7]=b[1]*b[6]-b[0]*b[7];
    m[8]=b[0]*b[4]-b[1]*b[3];
    return (1./(vectorielle(b.Lx,b.Ly)*b.Lz))*m;
}
float Matr9f::norme()const{
    float n=0;
    n+=(*this)[0]*(*this)[0];
    n+=(*this)[1]*(*this)[1];
    n+=(*this)[2]*(*this)[2];
    n+=(*this)[3]*(*this)[3];
    n+=(*this)[4]*(*this)[4];
    n+=(*this)[5]*(*this)[5];
    n+=(*this)[6]*(*this)[6];
    n+=(*this)[7]*(*this)[7];
    n+=(*this)[8]*(*this)[8];
    return sqrt(n);
}
//*
Matr9f pivotaxeangle(Vect3f const& axe, float teta){
    Matr9f m(axe.v());
    m = sin(teta)*m+(1-cos(teta))*m*m;
    m[0]+=1;
    m[4]+=1;
    m[8]+=1;
    return m;
}

void multglmatrix(Matr9f const& m){
    float m2[16] = //codage en colonne obligatoire...
    {
        m[0],m[3],m[6],0,
        m[1],m[4],m[7],0,
        m[2],m[5],m[8],0,
        0,0,0,1
    };
    glMultMatrixf(m2); //appel de la multiplication GL
}
void multgltmatrix(Matr9f const& m){
    float m2[16] = //codage en ligne car on transpose...
    {
        m[0],m[1],m[2],0,
        m[3],m[4],m[5],0,
        m[6],m[7],m[8],0,
        0,0,0,1
    };
    glMultMatrixf(m2); //appel de la multiplication GL
}
void Vect3f::print (const char* nom) const
    {printf("%s : %f %f %f \n",nom,x,y,z);}
void Matr9f::print (const char* nom) const
    {printf("%s\n",nom);Lx.print("");Ly.print("");Lz.print("");}


