#ifndef OCTREE_H
#define OCTREE_H


#define EMAX     30  //EXPOSANT maximal en 32 bit (à 31 on touche le bit de signe, à 32 on sort)
#define POSMAX    (1<<EMAX)


union int_float{//pour faire des opérations bit à bit sur des float
    int i;
    float f;
};
inline int getexposant(float valeur){union int_float dif; dif.f=valeur;return ((dif.i>>23)-126);}



/****************************************************************************
                            classe Octree
*****************************************************************************/

template<class T1>
class Octree{
    public:
    Octree<T1> *ptr;
    public:
    T1 node;

    Octree():node(){ptr=NULL;}//constructeur Octree vide
    Octree(Octree<T1> const &oct2):node(){// constructeur copie
        ptr=NULL;
        *this=oct2;
    }
    Octree(int n):node(){splitrec(n);}// constructeur allocation récursive


    ~Octree(){delete[] ptr;ptr=NULL;}// destructeur


    void operator=(Octree<T1> const &oct2){//copie récursive
        if(ptr!=NULL)delete[] ptr;
        ptr=NULL;
        node=oct2.node;
        if(oct2.ptr!=NULL){
            split();
            ptr[0]=oct2.ptr[0];
            ptr[1]=oct2.ptr[1];
            ptr[2]=oct2.ptr[2];
            ptr[3]=oct2.ptr[3];
            ptr[4]=oct2.ptr[4];
            ptr[5]=oct2.ptr[5];
            ptr[6]=oct2.ptr[6];
            ptr[7]=oct2.ptr[7];
        }
    }

    void splitrec(int n){//division récursive de l'Octree
        split();if(n>0){
        ptr[0].splitrec(n-1);
        ptr[1].splitrec(n-1);
        ptr[2].splitrec(n-1);
        ptr[3].splitrec(n-1);
        ptr[4].splitrec(n-1);
        ptr[5].splitrec(n-1);
        ptr[6].splitrec(n-1);
        ptr[7].splitrec(n-1);
        }
    }
    void split(){ptr = new Octree<T1>[8];}//division du cube
    void cut(){delete[] ptr;ptr=NULL;}// suppression des fils


    int compteextremite();// donne le nombre de feuilles de l'octree
};


#include "matr9f.h"


/****************************************************************************
                            classe Traceur
*****************************************************************************/

template<class T>
class Traceur{
        private:
    int posxyz[3];
    int e;
    Vect3f translation;
    float scale,_scale;//scale<<1     _scale>>1
    Octree<T>* tab[EMAX+2];
        public:

    // Les constructeurs
    inline Traceur(Octree<T>& oct0):translation(){
        tab[EMAX]=&oct0;
        tab[EMAX+1]=(Octree<T>*)(tab+EMAX);//la dernière case du tableau correspond à un cube qui pointe sur le cube racine
        e=EMAX;posxyz[0]=posxyz[1]=posxyz[2]=0;
        _scale=1<<e;scale=1./scale;
    }
    inline Traceur(Octree<T>& oct0, float echelle, Vect3f t):translation(t){
        tab[EMAX]=&oct0;
        tab[EMAX+1]=(Octree<T>*)(tab+EMAX);//la dernière case du tableau correspond à un cube qui pointe sur le cube racine
        e=EMAX;posxyz[0]=posxyz[1]=posxyz[2]=0;
        scale=echelle;_scale=1./echelle;
    }
    inline Traceur(Traceur<T> const&t){
        tab[EMAX]=t.tab[EMAX];
        tab[EMAX+1]=(Octree<T>*)(tab+EMAX);
        e=EMAX;posxyz[0]=posxyz[1]=posxyz[2]=0;
        scale=t.scale;_scale=t._scale;translation = t.translation;
    }

    // Accesseurs et mutateurs
    inline void setscale(float echelle){scale=echelle; _scale=1./echelle;}
    inline void setoffset(Vect3f const&u){translation=u;}
    inline float getscale()const{return scale;}
    inline float getscale_()const{return _scale;}
    inline Vect3f getoffset()const{return translation;}

    inline Octree<T>* courant()const{return tab[e];}
    inline Octree<T>* gethead()const{return tab[EMAX];}
    inline int getexp()const{return e;}
    inline float getlargeur()const{return scale*(1<<e);}
    inline const int* getposxyz()const{return posxyz;}
    inline Vect3f getcoordonne()const{return scale*Vect3f(posxyz[0],posxyz[1],posxyz[2])-translation;}
    // Coordonnées du coin du cube
    inline Vect3f getmiddlecoordonne()const{return scale*(Vect3f(posxyz[0],posxyz[1],posxyz[2])+(1<<e)*Vect3f(0.5,0.5,0.5))-translation;}
    // Coordonnées du milieu du cube
    inline T& getnode(int e0)const{return (*tab[e0]).node;}
    // avoir l'étiquette d'un père
    inline T& getnode()const{return (*tab[e]).node;}
    // avoir l'étiquette du cube courant


    inline int addposcube(int &e0);
    //incrémente le code de morton en cartésien, retourne l'indice et met l'exposant commun dans e0

    void lastcube( int posx2, int posy2 , int posz2);
    // Positionne le traceur au cube terminal de coordonnée demandée

    void getcube( int posx2,  int posy2 , int posz2,int e0);
    // Positionne le traceur au cube demandé si celui-ci existe,
    // sinon, a le même effet que "lastcube"

    inline bool plastcube( int posx2,  int posy2 , int posz2){if((posx2|posy2|posz2)>>EMAX !=0)return false; else lastcube(posx2,posy2,posz2);return true;}
    inline bool pgetcube( int posx2,  int posy2 , int posz2, int e0){if((posx2|posy2|posz2)>>EMAX !=0){return false; }else{getcube(posx2,posy2,posz2,e0);return e==e0;}}
    inline void lastcube(Vect3f u){u=_scale*(u+translation);  lastcube(u.x,u.y,u.z);}
    inline void getcube(Vect3f u,int e0){u=_scale*(u+translation);  getcube(u.x,u.y,u.z,e0);}
    inline bool plastcube(Vect3f u){u=_scale*(u+translation);  return plastcube(u.x,u.y,u.z);}
    inline bool pgetcube(Vect3f u,int e0){u=_scale*(u+translation);  return pgetcube(u.x,u.y,u.z,e0);}

    void firstlastcube();
    /*  ATTENTION : firstlastcube doit toujours être appelé lorsque
    l'on parcours l'arbre avec nextcube() sinon le parcours ne se fait pas */
    bool nextlastcube();
    void firstcube();
    bool nextcube();


    inline void split_go(){
        (*courant()).split();
        tab[e-1] = (*courant()).ptr;
        e--;
    }/* subdivise et se place au premier fils,
    ATTENTION : Il faut appeler cette fonction sur un cube terminal, et il faut
    que les coordonnées soient celles du fils,
    donc il vaut mieux utiliser cette fonction après "nextlastcube()" */

    Octree<T>* writeoct(int posx2, int posy2, int posz2, int e0);
    // si le cube n'existe pas à la position demandée, le cube est créé.

    inline Octree<T>* writeoct(Vect3f u, int e0){u=_scale*(u+translation);  return writeoct(u.x,u.y,u.z,e0);}
    inline T& writefeuille(int posx2,  int posy2,  int posz2, int e0){return (*writeoct(posx2,posy2,posz2,e0)).node;}
    inline T& writefeuille(Vect3f u, int e0){return (*writeoct(u, e0)).node;}
    inline Octree<T>* pwriteoct(int posx2, int posy2, int posz2, int e0){if((posx2|posy2|posz2)>>EMAX !=0){return NULL;}else{return writeoct(posx2,posy2,posz2,e0);}}
    inline Octree<T>* pwriteoct(Vect3f u, int e0){u=_scale*(u+translation);  return pwriteoct(u.x,u.y,u.z,e0);}


    int set8octzone(Vect3f coinzone,float largzone, Octree<T> **oct8);// renvoie l'exposant relatif aux 8 octree.
    /* donne dans le tableau oct8 qui doit déjà posséder 8 cases,
    les 8 Octree englobant la zone de largeur "largzone" et de coin "coinzone"*/

    void get8oct(Octree<T> **resultat,Vect3f milieu, int e0);
    /* donne dans le tableau resultat qui doit déjà posséder 8 cases,
    les 8 Octree entourant le vecteur "milieu" d'exposant "e0" */

};


/*****************************************************************
                Méthodes de la classe Traceur
******************************************************************/

template<class T>
inline int Traceur<T>::addposcube(int &e0){//retourne l'indice et met l'exposant commun dans e0
    int f1=posxyz[0] & posxyz[1] & posxyz[2];
    union int_float dif;
    dif.f=(float)((f1+(1<<e0))^f1);
    dif.f+=.5;// pour que l'exposant ne puisse pas être négatif
    int tc = (dif.i>>23)-127;//extraction de l'exposant
    int i =((posxyz[0]>>tc)&1)|(((posxyz[1]>>tc)&1)<<1)|(((posxyz[2]>>tc)&1)<<2);
    i++;
    posxyz[0]=(((posxyz[0]>>(tc+1))<<1)|(i&1))<<tc;
    posxyz[1]=(((posxyz[1]>>(tc+1))<<1)|((i&2)>>1))<<tc;
    posxyz[2]=(((posxyz[2]>>(tc+1))<<1)|((i&4)>>2))<<tc;
    e0=tc;
    return i;
}

template<class T> void Traceur<T>::lastcube(int posx2, int posy2 ,int posz2){
    unsigned int difference = (posxyz[0]^posx2)|(posxyz[1]^posy2)|(posxyz[2]^posz2);
    while(difference>>e!=0)e++;
    while((*tab[e]).ptr!=NULL){
        e--;
        int indice=((posx2>>e)&1)|(((posy2>>e)&1)<<1)|(((posz2>>e)&1)<<2);
        tab[e] = (*tab[e+1]).ptr+indice;
    }
    posxyz[0]=posx2;posxyz[1]=posy2;posxyz[2]=posz2;
}

template<class T> void Traceur<T>::getcube(int posx2, int posy2 , int posz2, int e0){
    unsigned int difference = (posxyz[0]^posx2)|(posxyz[1]^posy2)|(posxyz[2]^posz2);
    //if(difference>=POSMAX){printf("difference %d",difference);exit(41);}//debug line
    e=max(e,e0);
    while(difference>>e!=0)e++;
    while((*tab[e]).ptr!=NULL && e > e0){
        e--;
        int indice=((posx2>>e)&1)|(((posy2>>e)&1)<<1)|(((posz2>>e)&1)<<2);
        tab[e] = (*tab[e+1]).ptr+indice;;
    }
    posxyz[0]=posx2;posxyz[1]=posy2;posxyz[2]=posz2;
}

template<class T> void Traceur<T>::firstlastcube(){
    e=EMAX;
    posxyz[0]=posxyz[1]=posxyz[2]=0;
    while((*tab[e]).ptr!=NULL){
        e--;
        tab[e] = (*tab[e+1]).ptr;
    }
}
template<class T> bool Traceur<T>::nextlastcube(){
    int indice=addposcube(e);// On incrémente la position
    if(e>=EMAX){posxyz[0]=posxyz[1]=posxyz[2]=0;return false;}//On termine
    tab[e] = (*tab[e+1]).ptr+indice;
    while((*tab[e]).ptr!=NULL){// On prend le dernier cube
        e--;
        int indice=((posxyz[0]>>e)&1)|(((posxyz[1]>>e)&1)<<1)|(((posxyz[2]>>e)&1)<<2);
        tab[e] = (*tab[e+1]).ptr+indice;
    }
    return true;
}
template<class T> void Traceur<T>::firstcube(){e=EMAX;posxyz[0]=posxyz[1]=posxyz[2]=0;}
template<class T> bool Traceur<T>::nextcube(){
    if((*tab[e]).ptr!=NULL){//On va en profondeur
        e--;
        int indice=((posxyz[0]>>e)&1)|(((posxyz[1]>>e)&1)<<1)|(((posxyz[2]>>e)&1)<<2);
        tab[e] = (*tab[e+1]).ptr+indice;
    }else{//on remonte et on incrémente la position
        int indice=addposcube(e);
        if(e>=EMAX){posxyz[0]=posxyz[1]=posxyz[2]=0;return false;}//On termine
        tab[e] = (*tab[e+1]).ptr+indice;
    }
    return true;
}
template<class T>
Octree<T>* Traceur<T>::writeoct(int posx2, int posy2, int posz2, int e0){
    getcube(posx2,posy2,posz2,e0);
    while(e > e0){// si l'exposant n'est pas la bon, on subdivise jusqu'à l'atteindre
        (*courant()).split();
        e--;
        tab[e] = (*tab[e+1]).ptr + (((posx2>>e)&1)|(((posy2>>e)&1)<<1)|(((posz2>>e)&1)<<2));
    }
    return courant();
}

template<class T>
int Traceur<T>::set8octzone(Vect3f coinzone,float largzone, Octree<T> **oct8){
    int e0 = getexposant(_scale*largzone);
    oct8[0]=pwriteoct(coinzone,e0);
    oct8[1]=pwriteoct(coinzone+Vect3f(largzone,0,0),e0);
    oct8[2]=pwriteoct(coinzone+Vect3f(0,largzone,0),e0);
    oct8[3]=pwriteoct(coinzone+Vect3f(largzone,largzone,0),e0);
    oct8[4]=pwriteoct(coinzone+Vect3f(0,0,largzone),e0);
    oct8[5]=pwriteoct(coinzone+Vect3f(largzone,0,largzone),e0);
    oct8[6]=pwriteoct(coinzone+Vect3f(0,largzone,largzone),e0);
    oct8[7]=pwriteoct(coinzone+Vect3f(largzone,largzone,largzone),e0);
    if(oct8[0]==oct8[1])oct8[1]=NULL;//égalité en x
    if(oct8[2]==oct8[3])oct8[3]=NULL;
    if(oct8[4]==oct8[5])oct8[5]=NULL;
    if(oct8[6]==oct8[7])oct8[7]=NULL;

    if(oct8[0]==oct8[2])oct8[2]=NULL;//égalité en y
    if(oct8[1]==oct8[3])oct8[3]=NULL;
    if(oct8[4]==oct8[6])oct8[6]=NULL;
    if(oct8[5]==oct8[7])oct8[7]=NULL;

    if(oct8[0]==oct8[4])oct8[4]=NULL;//égalité en z
    if(oct8[1]==oct8[5])oct8[5]=NULL;
    if(oct8[2]==oct8[6])oct8[6]=NULL;
    if(oct8[3]==oct8[7])oct8[7]=NULL;
    return e0;
}
template <class T>
void Traceur<T>::get8oct(Octree<T> **resultat,Vect3f milieu, int e0){
    resultat[0]=resultat[1]=resultat[2]=resultat[3]=resultat[4]=resultat[5]=resultat[6]=resultat[7]=NULL;

    Vect3f u=_scale*(milieu+translation);
    //il est important de vite passer au coordonnées entière car les erreurs d'arrondi peuvent être dangereuses
    int p1=u.x,p2=u.y,p3=u.z;
    p1-=(1<<(e0-1));
    p2-=(1<<(e0-1));
    p3-=(1<<(e0-1));
    int l = 1<<e0;
    if(pgetcube(p1,p2,p3,e0))resultat[0]=courant();
    if(pgetcube(p1+l,p2,p3,e0))resultat[1]=courant();
    if(pgetcube(p1,p2+l,p3,e0))resultat[2]=courant();
    if(pgetcube(p1+l,p2+l,p3,e0))resultat[3]=courant();
    if(pgetcube(p1,p2,p3+l,e0))resultat[4]=courant();
    if(pgetcube(p1+l,p2,p3+l,e0))resultat[5]=courant();
    if(pgetcube(p1,p2+l,p3+l,e0))resultat[6]=courant();
    if(pgetcube(p1+l,p2+l,p3+l,e0))resultat[7]=courant();
}

/************** Une méthode de la classe Octree *****************/

template <class T1>
int Octree<T1>::compteextremite(){
    int indice=0;
    Traceur<T1> trac(*this);
    trac.firstlastcube();do{indice++;}while(trac.nextlastcube());
    return indice;
}


#endif
