#include "potentiel.h"

// constructeur avec  la transformation linéaire d'un autre potentiel de blobs
Potentiel::Potentiel(Potentiel const&p, Matr9f const&transformation):oct(),tab(),trac(oct,p.trac.getscale()*transformation.norme(),transformation.norme()*p.trac.getoffset()){
    int k;
    for(k=p.tab.end()-1;k>=0;k--)addblob(Blob(transformation*p.tab(k).pos,!transformation*p.tab(k).influence));
}

// insère un blob dans le tableau et l'octree
void Potentiel::addblob(Blob const&b){
    tab.add(b);
    OctreeIndices *oct8[8];
    trac.set8octzone(b.pos-b.getlargeurzone()*Vect3f(1,1,1),b.getlargeurzone()*2,oct8);
    int i;
    for(i=7;i>=0;i--)
        if(oct8[i]!=NULL)(*oct8[i]).node.add(tab.end()-1);
}

// clacul du potentiel avec seulement les blobs proches du points
float Potentiel::operator()(Vect3f const&a){
    float p=0;
    if(trac.plastcube(a)){
        int i,j;
        for(j=trac.getexp();j<=EMAX;j++){
            Indicestab &t=trac.getnode(j);
            for(i=0;i<t.end();i++)p+=tab(t(i))(a);
        }
    }
    return p;
}

// méthode de newton pour plaquer le vecteur contre la surface
void Pot0::plaquevect(Vect3f &pos, float ref){
    float p0 = (*this)(pos);
    int i;
    for(i=0;i<10 && fabs(p0-ref)>0.000001;i++){
        Vect3f grad = PRECISION_DERIVEE_1*Vect3f((*this)(pos+Vect3f(PRECISION_DERIVEE,0,0))-p0,
                                    (*this)(pos+Vect3f(0,PRECISION_DERIVEE,0))-p0,
                                    (*this)(pos+Vect3f(0,0,PRECISION_DERIVEE))-p0);
        pos+=(ref-p0)/(grad*grad)*grad;
        p0 = (*this)(pos);
    }
}

//ajoute un blob coloré
void Couleurs::addcouleur(Blob const&b, int couleur){
    tab.add(Zone(b,couleur2Vect3f(couleur)));
    OctreeIndices *oct8[8];
    trac.set8octzone(b.pos-b.getlargeurzone()*Vect3f(1,1,1),b.getlargeurzone()*2,oct8);
    int i;
    for(i=7;i>=0;i--)
        if(oct8[i]!=NULL)(*oct8[i]).node.add(tab.end()-1);
}

//récupère la couleur à une position donnée
int Couleurs::operator()(Vect3f const&a){
    float cardinal=0;
    Vect3f var;
    if(trac.plastcube(a)){
        int i,j;
        for(j=trac.getexp();j<=EMAX;j++){
            Indicestab &t=trac.getnode(j);
            for(i=0;i<t.end();i++){
                float poid=tab(t(i)).b(a);
                cardinal+=poid;
                var+=poid*tab(t(i)).a;
            }
        }
    }
    if(cardinal==0)return 0;
    else return Vect3f2couleur((1./cardinal)*var);
}

